/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * input.c
 */

/*X
   Module name: input.c
   X */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/AsciiText.h>

#include "calc.h"
#include "../xstuff/fonts.h"

#define ABORT	0
#define	DONE	1
#define	WAITING	2

Widget pshell, dialog, dialogLabel, dialogText, dialogCancel, dialogDone;

int status;

static void DialogDone(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(pshell);
	status = DONE;
}

static void DialogAbort(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtPopdown(pshell);
	status = ABORT;
}

void DialogCancelAction(Widget w, XEvent * event, String * params, Cardinal * n)
{
	XtPopdown(pshell);
	status = ABORT;
}

void DialogDoneAction(Widget w, XEvent * event, String * params, Cardinal * n)
{
	XtPopdown(pshell);
	status = DONE;
}

void init_input()
{
	pshell = XtVaCreatePopupShell("pshell",
		transientShellWidgetClass, topLevel, NULL);
	dialog = XtVaCreateManagedWidget("dialog",
		formWidgetClass, pshell, NULL);
	dialogLabel = XtVaCreateManagedWidget("dialogLabel",
		labelWidgetClass, dialog, NULL);
	dialogText = XtVaCreateManagedWidget("dialogText",
		asciiTextWidgetClass, dialog,
		XtNeditType, XawtextEdit, NULL);
	dialogDone = XtVaCreateManagedWidget("dialogDone",
		commandWidgetClass, dialog, NULL);
	dialogCancel = XtVaCreateManagedWidget("dialogCancel",
		commandWidgetClass, dialog, NULL);

	XtAddCallback(dialogDone, XtNcallback, DialogDone, NULL);
	XtAddCallback(dialogCancel, XtNcallback, DialogAbort, NULL);
}

int ask_for_str_comp(char *prompt, char *buffr, int (*comp) ())
{
	Position x, y;
	Dimension width, height, w1, h1;
	String string;

	status = WAITING;

	XtVaGetValues(topLevel,
		XtNwidth, &width,
		XtNheight, &height, NULL);
	XtVaGetValues(pshell,
		XtNwidth, &w1,
		XtNheight, &h1, NULL);
	XtTranslateCoords(topLevel,
	       (Position) (width - w1) / 2, (Position) (height - h1) / 2,
		&x, &y);
	XtVaSetValues(pshell,
		XtNx, x,
		XtNy, y, NULL);
	XtVaSetValues(dialogLabel,
		XtNlabel, prompt,
		XtNwidth, w1-10, NULL);
	XtVaSetValues(dialogText,
		XtNstring, buffr,
		XtNinsertPosition, strlen(buffr), NULL);
	XtPopup(pshell, XtGrabNonexclusive);

	/* This sucks big time. We don't want to return until the
	   user has done something, but that also means that
	   we won't return to XtAppMainLoop, so no events will
	   be processed. The only way to avoid deadlock is to
	   make this ugly busy-wait implement the XtAppMainLoop
	   functionality. That's what you get from squeezing
	   Xlib applications into an Xt framework: doesn't fit.
	 */
	while (status == WAITING) {
		XEvent event_return;

		XtAppNextEvent(app_context, &event_return);
		XtDispatchEvent(&event_return);
	}

	XtVaGetValues(dialogText,
		XtNstring, &string, NULL);
	strcpy(buffr, string);
	return status;
}

/*X
   static int nocomp(char *b)
   This particular completion function doesn't complete at all, it just
   returns TRUE, making TAB equivalent to RET and LFD.
   X */
static int nocomp(char *b)
{
	return TRUE;
}

/*X
   int ask_for_str(char *prompt, char *buffr)
   Calls ask_for_str_comp with nocomp as completion function.
   X */
/* 95-06-29: changed "buffer" to "buffr" to please gcc */
int ask_for_str(char *prompt, char *buffr)
{
	return ask_for_str_comp(prompt, buffr, nocomp);
}

