/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>
#include <sys/stat.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Paned.h>

#include "types.h"
#include "calc.h"

#include "bitmaps/new.xbm"
#include "bitmaps/fld_open.xbm"
#include "bitmaps/floppy3.xbm"
#include "bitmaps/preview.xbm"
#include "bitmaps/printer.xbm"
#include "bitmaps/plotter.xbm"
#include "bitmaps/info.xbm"
#include "bitmaps/copyright.xbm"

#ifndef SIAGHOME
#define SIAGHOME "/usr/local/siag"
#endif
#ifndef SIAGHELP
#define SIAGHELP "chimera"
#endif
#ifndef SIAGRC
#define SIAGRC ".siagrc"
#endif

/* siagmenu.c */

XtAppContext app_context;

Widget topLevel, topbox, box, form, gridpane,
	/*viewport,*/ text1, label1, label2/*,
	selectall, colnum, rownum, grid, hscroll, vscroll*/;

/* The toolbar */
Widget tbNew, tbOpen, tbSave, tbView, tbPrint, tbPlot, tbHelp, tbCopy;

char *version = "Siag (Scheme In A Grid) 0.97. No Warranty.";
char *siaghome, *siaghelp;
static char *siagrc;

int /*loopflag = 1, */ok2print = 0;

extern void print_cells();

static void set_point(int row, int col)
{
	position p;

	p.row = row;
	p.col = col;
	move_point(w_list, p);
}

static void set_mark(int row, int col)
{
	position p;

	p.row = row;
	p.col = col;
	move_mark(w_list, p);
}

static void set_block(int row1, int col1, int row2, int col2)
{
	if (row1 < row2) {
		w_list->blku.row = row1;
		w_list->blkl.row = row2;
	} else {
		w_list->blku.row = row2;
		w_list->blkl.row = row1;
	}
	if (col1 < col2) {
		w_list->blku.col = col1;
		w_list->blkl.col = col2;
	} else {
		w_list->blku.col = col2;
		w_list->blkl.col = col1;
	}
}

void ColnumButtonAction(Widget w, XEvent * event, String * params, Cardinal * n)
{
	int col, row, x, y;

	x = event->xbutton.x;
	y = 0;
	hide_cur(w_list);
	w_list = find_window_by_widget(w);
	get_coords_cell(w_list, w_list->top.row, w_list->top.col, 
			&row, &col, x, y);
	if (*n < 1 || !strcmp(params[0], "set"))
		/* Set block to this column */
		set_block(1, col, BUFFER_ROWS, col);
	else if (!strcmp(params[0], "left"))
		w_list->blku.col = col;
	else if (!strcmp(params[0], "right"))
		w_list->blkl.col = col;
	pr_scr_flag = TRUE;
	show_cur(w_list);
}

void RownumButtonAction(Widget w, XEvent * event, String * params, Cardinal * n)
{
	int col, row, x, y;

	x = 0;
	y = event->xbutton.y;
	hide_cur(w_list);
	w_list = find_window_by_widget(w);
	get_coords_cell(w_list, w_list->top.row, w_list->top.col, 
			&row, &col, x, y);
	if (*n < 1 || !strcmp(params[0], "set"))
		/* Set block to this row */
		set_block(row, 1, row, BUFFER_COLS);
	else if (!strcmp(params[0], "top"))
		w_list->blku.row = row;
	else if (!strcmp(params[0], "bottom"))
		w_list->blkl.row = row;
	pr_scr_flag = TRUE;
	show_cur(w_list);
}

void GridButtonAction(Widget w, XEvent * event, String * params, Cardinal * n)
{
	int col, row;
	int x, y;

	x = event->xbutton.x;
	y = event->xbutton.y;
	hide_cur(w_list);
	w_list = find_window_by_widget(w);
	get_coords_cell(w_list, w_list->top.row, w_list->top.col, 
			&row, &col, x, y);
	if (*n < 1 || !strcmp(params[0], "point"))
		set_point(row, col);
	else if (!strcmp(params[0], "mark"))
		set_mark(row, col);
	else if (!strcmp(params[0], "block")) {
		set_block(row, col,
			mark_position(w_list).row, mark_position(w_list).col);
		pr_scr_flag = TRUE;
	}
	show_cur(w_list);
}

static int oldmain(argc, argv)
int argc;
char **argv;
{
	int count = 0;
	spread **matrix;
	buffer *b = NULL;
	char path[FILENAME_MAX];

	if (init_pool()) {
		fprintf(stderr, "Couldn't initialize string pool\n");
		exit(1);
	}
	init_parser(argc, argv);
	init_mathwrap();	/* standard FP library functions */

	if (count == 0) {	/* default to "noname" */
		matrix = new_matrix();
		calc_matrix(matrix);
		b = new_buffer(matrix, "noname", "noname");
	}
	init_windows(b, argc, argv);

	init_input();
	font_init(topLevel);
	init_cmds();
	while (++argv, --argc) {
		if (**argv != '-') {
			matrix = new_matrix();
			strcpy(path, *argv);
			b = new_buffer(matrix, buffer_name(*argv), path);
			loadmatrix(path, b);
			calc_matrix(matrix);
			b->change = FALSE;
			w_list->buf = b;
			count++;
		}
	}
	pr_scr_flag = TRUE;
	ok2print = 1;
	return 0;
}

static void pr_colnum(Widget w, XEvent * event,
		      String * params, Cardinal * num_params)
{
	if (ok2print)
		draw_colnums(display, find_window_by_widget(w));
}

static void pr_rownum(Widget w, XEvent * event,
		      String * params, Cardinal * num_params)
{
	if (ok2print)
		draw_rownums(display, find_window_by_widget(w));
}

static void pr_grid(Widget w, XEvent * event,
		    String * params, Cardinal * num_params)
{
	if (ok2print) {
		draw_cells(display, find_window_by_widget(w));
		show_cur(w_list);
	}
}

#if 0
static void newgrid(Widget w, XtPointer client_data, XtPointer call_data)
{
	XtVaCreateManagedWidget("NewGrid",
		labelWidgetClass, gridpane, NULL);
}
#endif

static void execute_action(Widget w, XEvent * event,
			   String * params, Cardinal * num_params)
{
	char b[256];
	int i;

	strcpy(b, "(");
	strcat(b, params[0]);
	for (i = 1; i < *num_params; i++) {
		strcat(b, " ");
		strcat(b, params[i]);
	}
	strcat(b, ")");
	execute(b);
}

static void
helptext_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	char b[256];
	int i;

	strcpy(b, params[0]);
	for (i = 1; i < *num_params; i++) {
		strcat(b, " ");
		strcat(b, params[i]);
	}
	XtVaSetValues(label1, XtNlabel, b, NULL);
}

static void draw_input_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	char *p = ret_taddr(w_list->buf->matrix,
		w_list->point_pos.row, w_list->point_pos.col);
	if (p == NULL)
		p = "";
	draw_input(display, p);
}

static void
siaghelp_action(Widget w, XEvent *event,
	String *params, Cardinal *num_params)
{
	char b[256];

	sprintf(b, "file://localhost%s/siag/docs/%s", siaghome, params[0]);
	if (!fork()) {
		execlp(siaghelp, "Siaghelp", "-geometry", "523x356", b, NULL);
		exit(0);
	}
}

static XtActionsRec actions[] =
{
	{"pr-colnum", pr_colnum},
	{"pr-rownum", pr_rownum},
	{"pr-grid", pr_grid},
	{"grid-button", GridButtonAction},
	{"colnum-button", ColnumButtonAction},
	{"rownum-button", RownumButtonAction},
	{"dialog-done", DialogDoneAction},
	{"dialog-cancel", DialogCancelAction},
	{"execute", execute_action},
	{"helptext", helptext_action},
	{"draw-input", draw_input_action},
	{"siaghelp", siaghelp_action}
};

static void execute_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
	char *b = (char *) client_data;

	execute(b);
}

typedef struct _menu_struct {
	String name;
	XtPointer function;
} menu_struct;

static menu_struct mnuFile_names[] =
{
	{"mnuFileNew", "(new-siag)"},
	{"mnuFileOpen", "(load-buffer)"},
	{"mnuFileSave", "(save-buffer)"},
	{"mnuFileSaveAs", "(save-buffer-as)"},
	{"mnuFilePreview", "(preview-block)"},
	{"mnuFilePrint", "(print-block)"},
	{"-", NULL},
	{"mnuFileExit", "(quit-siag)"},
	{NULL, NULL}
};

static menu_struct mnuEdit_names[] =
{
	{"mnuEditDelete", "(delete-cell)"},
	{"mnuEditInsertLine", "(insert-line)"},
	{"-", NULL},
	{"mnuEditLabel", "(edit-label)"},
	{"mnuEditExpression", "(edit-expression)"},
	{NULL, NULL}
};

static menu_struct mnuBlock_names[] =
{
	{"mnuBlockSetBlock", "(set-block)"},
	{"mnuBlockSetMark", "(set-mark-command)"},
	{"mnuBlockCopyBlock", "(copy-block)"},
	{"mnuBlockDeleteBlock", "(delete-block)"},
	{"-", NULL},
	{"mnuBlockBorders", "(block-borders 1)"},
	{"mnuBlockGrid", "(block-borders 2)"},
	{"mnuBlockNone", "(block-borders 0)"},
	{NULL, NULL}
};

static menu_struct mnuFormat_names[] =
{
	{"mnuFormatColumnWidth", "(set-cell-width)"},
	{"mnuFormatRowHeight", "(set-cell-height)"},
	{"-", NULL},
	{"mnuFormatCellFormat", "(set-cell-format)"},
	{"mnuFormatBlockFormat", "(set-block-format)"},
	{NULL, NULL}
};

static menu_struct mnuBuffer_names[] =
{
	{"mnuBufferChange", "(switch-to-buffer)"},
	{"mnuBufferKill", "(kill-buffer)"},
	{"-", NULL},
	{"mnuWindowSplit", "(split-window-vertically)"},
	{"mnuWindowDelete", "(delete-window)"},
	{"mnuWindowMerge", "(delete-other-windows)"},
	{"mnuWindowOther", "(other-window)"},
	{NULL, NULL}
};

static menu_struct mnuPlot_names[] =
{
	{"mnuPlotLines", "(plot-block 0)"},
	{"mnuPlotPoints", "(plot-block 1)"},
	{"mnuPlotLinespoints", "(plot-block 2)"},
	{"mnuPlotImpulses", "(plot-block 3)"},
	{"mnuPlotDots", "(plot-block 4)"},
	{"mnuPlotSteps", "(plot-block 6)"},
	{"mnuPlotBoxes", "(plot-block 7)"},
	{"-", NULL},
	{"mnuPlotAdvanced", "(no-op)"},
	{NULL, NULL}
};

static menu_struct mnuHelp_names[] =
{
	{"mnuHelpContents", "(help-contents)"},
	{"mnuHelpSearch", "(help-search)"},
	{"mnuHelpCopying", "(help-copyright)"},
	{"mnuHelpForHelp", "(help-for-help)"},
	{"-", NULL},
	{"mnuHelpAboutSiag", "(print-version)"},
	{NULL, NULL}
};

void vscroll_jump(Widget w, XtPointer client_data, XtPointer call_data)
{
	float top;
	int gridtop;

	hide_cur(w_list);
	w_list = find_window_by_widget(w);
	XtVaGetValues(w, XtNtopOfThumb, &top, NULL);
	gridtop = top*BUFFER_ROWS;
	if (gridtop < 1) gridtop = 1;
	w_list->point_pos.row = w_list->top.row = gridtop;
	pr_scr_flag = TRUE;
	show_cur(w_list);
}

void vscroll_scroll(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i = (int) call_data;
	Dimension length;

	w_list = find_window_by_widget(w);
	XtVaGetValues(w, XtNlength, &length, NULL);
	if (i < 0)
		if ((length / -i) > 15)
			execute("(scroll-cell-up)");
		else
			execute("(scroll-up)");
	else
		if ((length / i) > 15)
			execute("(scroll-cell-down)");
		else
			execute("(scroll-down)");
}

void hscroll_jump(Widget w, XtPointer client_data, XtPointer call_data)
{
	float top;
	int gridtop;

	hide_cur(w_list);
	w_list = find_window_by_widget(w);
	XtVaGetValues(w, XtNtopOfThumb, &top, NULL);
	gridtop = top*BUFFER_COLS;
	if (gridtop < 1) gridtop = 1;
	w_list->point_pos.col = w_list->top.col = gridtop;
	pr_scr_flag = TRUE;
	show_cur(w_list);
}

void hscroll_scroll(Widget w, XtPointer client_data, XtPointer call_data)
{
	int i = (int) call_data;
	Dimension length;

	w_list = find_window_by_widget(w);
	XtVaGetValues(w, XtNlength, &length, NULL);
	if (i < 0)
		if ((length / -i) > 15)
			execute("(scroll-cell-left)");
		else
			execute("(scroll-left)");
	else
		if ((length / i) > 15)
			execute("(scroll-cell-right)");
		else
			execute("(scroll-right)");
}

static void make_menu(menu_struct menu_entry_names[], Widget menu)
{
	int i;
	Widget entry;

	for (i = 0; menu_entry_names[i].name; i++) {
		String item = menu_entry_names[i].name;

		if (item[0] == '-')	/* line pane */
			entry = XtCreateManagedWidget(item,
						smeLineObjectClass, menu,
						      NULL, 0);
		else {
			entry = XtCreateManagedWidget(item,
						 smeBSBObjectClass, menu,
						      NULL, 0);
			XtAddCallback(entry,
				      XtNcallback, execute_callback,
				      menu_entry_names[i].function);
		}
	}
}

static Widget make_toolbar(char *name, char *cmd, Widget pw, 
	char *bits, int width, int height)
{
	Pixel fg, bg;
	Display *display = XtDisplay(pw);

	Widget w = XtVaCreateManagedWidget(name,
		commandWidgetClass, pw, NULL);
	XtVaGetValues(w,
		XtNforeground, &fg,
		XtNbackground, &bg, NULL);
	XtVaSetValues(w,
		XtNbitmap,
		XCreatePixmapFromBitmapData(display,
			XDefaultRootWindow(display),
			bits, width, height, fg, bg,
			XDefaultDepthOfScreen(XtScreen(w))), NULL);
	XtAddCallback(w,
		XtNcallback, execute_callback, (XtPointer)cmd);
	return w;
}

int main(int argc, char **argv)
{
	char b[256], *home;
	struct stat statbuf;

	Widget btnFile, mnuFile, btnEdit, mnuEdit, btnBlock, mnuBlock,
		btnFormat, mnuFormat, btnBuffer, mnuBuffer,
		btnPlot, mnuPlot, btnHelp, mnuHelp;

	if ((siaghome = getenv("SIAGHOME")) == NULL)
		siaghome = SIAGHOME;
	if ((siaghelp = getenv("SIAGHELP")) == NULL)
		siaghelp = SIAGHELP;
	if ((home = getenv("HOME")) == NULL)
		home = ".";
	sprintf(b, "%s/%s", home, SIAGRC);
	siagrc = strdup(b);

	XtSetLanguageProc(NULL, (XtLanguageProc) NULL, NULL);

	topLevel = XtVaAppInitialize(
		    &app_context,	/* application context */
		    "Siag",		/* application class */
		    NULL, 0,		/* command line options list */
		    &argc, argv,	/* command line args */
		    NULL,		/* for missing app-defaults file */
		    NULL);		/* terminate varargs list */

	XtAppAddActions(app_context, actions, XtNumber(actions));

	topbox = XtCreateManagedWidget("topbox",
		formWidgetClass, topLevel, NULL, 0);

	box = XtCreateManagedWidget("box",
		formWidgetClass, topbox, NULL, 0);

	btnFile = XtVaCreateManagedWidget("btnFile",
		menuButtonWidgetClass, box, NULL);
	mnuFile = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnFile, NULL);
	make_menu(mnuFile_names, mnuFile);

	btnEdit = XtVaCreateManagedWidget("btnEdit",
		menuButtonWidgetClass, box, NULL);
	mnuEdit = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnEdit, NULL);
	make_menu(mnuEdit_names, mnuEdit);

	btnBlock = XtVaCreateManagedWidget("btnBlock",
		menuButtonWidgetClass, box, NULL);
	mnuBlock = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnBlock, NULL);
	make_menu(mnuBlock_names, mnuBlock);

	btnFormat = XtVaCreateManagedWidget("btnFormat",
		menuButtonWidgetClass, box, NULL);
	mnuFormat = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnFormat, NULL);
	make_menu(mnuFormat_names, mnuFormat);

	btnBuffer = XtVaCreateManagedWidget("btnBuffer",
		menuButtonWidgetClass, box, NULL);
	mnuBuffer = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnBuffer, NULL);
	make_menu(mnuBuffer_names, mnuBuffer);

	btnPlot = XtVaCreateManagedWidget("btnPlot",
		menuButtonWidgetClass, box, NULL);
	mnuPlot = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnPlot, NULL);
	make_menu(mnuPlot_names, mnuPlot);

	btnHelp = XtVaCreateManagedWidget("btnHelp",
		menuButtonWidgetClass, box, NULL);
	mnuHelp = XtVaCreatePopupShell("menu",
		simpleMenuWidgetClass, btnHelp, NULL);
	make_menu(mnuHelp_names, mnuHelp);

	tbNew = make_toolbar("tbNew", "(new-siag)", box,
		new_bits, new_width, new_height);
	tbOpen = make_toolbar("tbOpen", "(load-buffer)", box,
		fld_open_bits, fld_open_width, fld_open_height);
	tbSave = make_toolbar("tbSave", "(save-buffer-as)", box,
		floppy3_bits, floppy3_width, floppy3_height);
	tbView = make_toolbar("tbView", "(preview-block)", box,
		preview_bits, preview_width, preview_height);
	tbPrint = make_toolbar("tbPrint", "(print-block)", box,
		printer_bits, printer_width, printer_height);
	tbPlot = make_toolbar("tbPlot", "(plot-block 0)", box,
		plotter_bits, plotter_width, plotter_height);
	tbHelp = make_toolbar("tbHelp", "(help-contents)", box,
		info_bits, info_width, info_height);
	tbCopy = make_toolbar("tbCopy", "(help-copyright)", box,
		copyright_bits, copyright_width, copyright_height);
		
	label1 = XtVaCreateManagedWidget("label1",
		labelWidgetClass, topbox, NULL);

	gridpane = XtVaCreateManagedWidget("gridpane",
		panedWidgetClass, topbox, NULL);
#if 0
	viewport = XtVaCreateManagedWidget("viewport",
		formWidgetClass, gridpane, NULL);

	selectall = XtVaCreateManagedWidget("selectall",
		commandWidgetClass, viewport, NULL);
	XtAddCallback(selectall, XtNcallback, newgrid, NULL);

	colnum = XtVaCreateManagedWidget("colnum",
		coreWidgetClass, viewport, NULL);

	rownum = XtVaCreateManagedWidget("rownum",
		coreWidgetClass, viewport, NULL);

	grid = XtVaCreateManagedWidget("grid",
		coreWidgetClass, viewport, NULL);

	vscroll = XtVaCreateManagedWidget("vscroll",
		scrollbarWidgetClass, viewport, NULL);

	hscroll = XtVaCreateManagedWidget("hscroll",
		scrollbarWidgetClass, viewport, NULL);
#endif
	label2 = XtVaCreateManagedWidget("label2",
		labelWidgetClass, topbox, NULL);
#if 0
	XtAddCallback(vscroll, XtNjumpProc, vscroll_jump, NULL);
	XtAddCallback(vscroll, XtNscrollProc, vscroll_scroll, NULL);

	XtAddCallback(hscroll, XtNjumpProc, hscroll_jump, NULL);
	XtAddCallback(hscroll, XtNscrollProc, hscroll_scroll, NULL);
#endif
	oldmain(argc, argv);
/*	XtRealizeWidget(topLevel);*/
	fsel_init(topLevel);

	sprintf(b, "(load \"%s/siag/siag.scm\")", siaghome);
	execute(b);

	if (!stat(siagrc, &statbuf)) {
		sprintf(b, "(load \"%s\")", siagrc);
		execute(b);
	}

	execute("(print-version)");

	XtAppMainLoop(app_context);

	return 0;
}

